import unittest
from myFitter import *
from myFitter.fit import _get_basis
from math import log, pi, sqrt

def relDev(a, b):
    return 0.0 if a == b == 0.0 else 2.0*(a-b)/(abs(a)+abs(b))

class MyModel(Model):
    def __init__(self):
        Model.__init__(self)

        self.addParameter('p1')
        self.addParameter('p2')

        self.addGaussianCSC('o1', 0.0, 1.0)
        self.addGaussianCSC('o2', 0.0, 1.0)
        self.addGaussianCSC('o3', 1.0, 1.0)

    def getDependentQuantities(self, p):
        q = Model.getDependentQuantities(self, p)

        q['o1'] = p['p1']
        q['o2'] = p['p2']
        q['o3'] = -p['p1']**2 - p['p2']**2
        q['p1p2'] = p['p1']*p['p2']

        return q


class FitTest(unittest.TestCase):
    def test_fit(self):
        model = MyModel()

        r = fit(model, start={'p1':1.0, 'p2':1.0}, solvercfg={'acc' : 1e-6})
        self.assertTrue(r.success)
        self.assertAlmostEqual(relDev(r.minChiSquare - 3.0*log(2.0*pi), 1.0), 0.0, places=5)
        self.assertAlmostEqual(r.bestFitParameters['p1'], 0.0, places=4)
        self.assertAlmostEqual(r.bestFitParameters['p2'], 0.0, places=4)

        model.par['p1'].lowerBound = 1.0
        model.par['p1'].upperBound = 2.0
        r = fit(model, start={'p1':1.5, 'p2':1.0}, solvercfg={'acc' : 1e-6})
        self.assertTrue(r.success)
        self.assertAlmostEqual(relDev(r.minChiSquare - 3.0*log(2.0*pi), 5.0), 0.0, places=5)
        self.assertAlmostEqual(r.bestFitParameters['p1'], 1.0, places=5)
        self.assertAlmostEqual(r.bestFitParameters['p2'], 0.0, places=5)

        r = fit(model, start={'p1':1.5, 'p2':1.0},
                basis=[{'p1':-0.3, 'p2':1.4}, {'p1':2.0, 'p2':-0.9}],
                solvercfg={'acc' : 1e-6})
        self.assertTrue(r.success)
        self.assertAlmostEqual(relDev(r.minChiSquare - 3.0*log(2.0*pi), 5.0), 0.0, places=5)
        self.assertAlmostEqual(r.bestFitParameters['p1'], 1.0, places=5)
        self.assertAlmostEqual(r.bestFitParameters['p2'], 0.0, places=5)
        
        model.par['p1'].lowerBound = None
        model.con['hyp'] = Constraint('p1p2', '>', 1.0)
        r = fit(model, start={'p1':1.5, 'p2':1.0}, solvercfg={'acc' : 1e-6})
        self.assertTrue(r.success)
        self.assertAlmostEqual(relDev(r.minChiSquare - 3.0*log(2.0*pi), 11.0), 0.0, places=5)
        self.assertAlmostEqual(r.bestFitParameters['p1'], 1.0, places=5)
        self.assertAlmostEqual(r.bestFitParameters['p2'], 1.0, places=5)

        r = fit(model, start={'p1':15, 'p2':10},
                basis=[{'p1':-15.0, 'p2':70.0},
                       {'p1':100.0, 'p2':-45.0}],
                solvercfg={'acc' : 1e-8})
        self.assertTrue(r.success)
        self.assertAlmostEqual(relDev(r.minChiSquare - 3.0*log(2.0*pi), 11.0), 0.0, places=5)
        self.assertAlmostEqual(r.bestFitParameters['p1'], 1.0, places=5)
        self.assertAlmostEqual(r.bestFitParameters['p2'], 1.0, places=5)

        model.con['hyp'] = Constraint('p1p2', '=', 1.0)
        r = fit(model, start={'p1':1.5, 'p2':1.0},
                solvercfg={'acc' : 1e-6})
        self.assertTrue(r.success)
        self.assertAlmostEqual(relDev(r.minChiSquare - 3.0*log(2.0*pi), 11.0), 0.0, places=5)
        self.assertAlmostEqual(r.bestFitParameters['p1'], 1.0, places=5)
        self.assertAlmostEqual(r.bestFitParameters['p2'], 1.0, places=5)

        r = fit(model, start={'p1':1.5, 'p2':1.0},
                solvercfg={'acc' : 1e-6, 'maxiter' : 1})
        self.assertFalse(r.success)
        self.assertEqual(r.status, 9)

        
